function [x, Nv, AFPar] = ReadAudioData(varargin)
% function [x, Nv, AFPar] = ReadAudioData(AFPar, offs, Ns)
%   x = ReadAudioData(File)
%   [x, Nv, AFPar] = ReadAudioData(File, offs, Ns)
%   [x, Nv] = ReadAudioData(File, offs, Ns)
%   [x, Nv, AFPar] = ReadAudioData(AFPar, offs, Ns)
%
%   x     <-  Audio data (double-precision values), in the nominal range -1 to
%             +1 for audio files containing integer data. The output data is
%             dimensioned Ns by M, where M is the number of channels.
%   Nv    <-  Number of samples (not including trailing zeros) returned. This
%             value will be less than Ns if an end-of-file has been reached.
%   AFPar <-  Output structure containing information about the file. The
%             presence of this ouput argument enables the buffered mode of
%             operation.
%   AFPar  -> Structure containing information about the file, including a
%             buffer of samples, if appropriate. The audio file name is part of
%             this structure. The first input argument must be either AFPar or
%             Fname.
%   Fname  -> File name (on the Matlab path, or with an explicit path). On
%             Windows, the file name is not case sensitive. The first input
%             argument must be either AFPar or Fname.
%   offs   -> Sample offset from the beginning of the data in the file (default
%             zero, beginning of the file). If the offset offs is less than
%             zero, zeros are added ahead of the data from the file.
%   Ns     -> Number of samples to be read (default to end-of-file). If the read
%             operation extends beyond the end-of-file, zeros are appended to
%             give an output with Ns samples per channel.
%
% Read data from an audio file. This routine handles several audio file types
% using the audioread function. In one mode, this routine maintains an internal
% buffer to make file access more efficient when accessing small chunks of data.
% In this routine the data in the file is treated as if it were preceded and
% followed with zeros.
%
% This routine uses the Matlab routine audioread to access the data. File types
% supported are documented in the description of audioread.
%
% Unbuffered mode:
%   x = ReadAudioData(File)
%   [x, Nv] = ReadAudioData(File, offs, Ns)
%   AFp = PrintAudioPar(File)
%   [x, Nv] = ReadAudioData(AFPar, offs, Ns)
% The unbuffered mode is a simple interface for reading data from an audio file.
% The first example above reads all of the data from an audio file. The second
% example reads a chunk of Ns samples starting at offset offs (0 is the start of
% the data in the file). In the first two  examples, PrintAudioPar is called by
% ReadAudioData to print the audio file parameters.The third example uses AFPar
% as obtained from PrintAudioPar. This allows one to check the parameters of the
% audio file before calling ReadAudioData.
%
% Buffered mode:
%   AFPar = PrintAudioPar(File)
%   [x, Nv, AFPar] = ReadAudioData(AFPar, offs, Ns)
% This case is used for repeated reading of partial data from an audio file. The
% file should be accessed first with PrintAudioPar which provides the structure
% AFPar which is used as input (and output) for ReadAudioData.
%
% If the output structure variable AFPar is used, the input structure AFPar is
% updated to contain a buffer of samples. For partial file reads, the buffer is
% updated to include read-ahead data to allow subsequent requests to get data
% without accessing the file.
%
% An example of the use of ReadAudioData to read a sequence of blocks of data
% with NBlock samples, with the block advancing by NAdv samples. If NAdv<NBlock,
% the blocks of data have overlapping samples.
%   AFPar = PrintAudioPar(FileName);
%   NBlock = 240;
%   NAdv = 180;
%   ioffs = 0;
%   while (1)
%     [x, Nv, AFPar] = ReadAudioData(AFPar, ioffs, NBlock);
%     if (Nv < NBlock)
%       break;
%     end
%     ... % Process a block of NBlock samples
%     ioffs = ioffs + NAdv;
%   end
% In this example, processing stops if a block contains less than NBlock samples
% from the file.
%
% Notes:
% - The structure AFPar can be cleared manually ("clear AFPar") to recover the
%   space used for buffering.
% - The maximum buffer size can be set manually by defining a value for
%   AFPar.NBuffMax. The default is 50000 samples per channel. The file does not
%   not need to be accessed again to refresh the buffer unless not all of the
%   requested samples are available in the buffer.
% - This routine throws an error if the input file contains A-law or mu-law
%   data and the Matlab version is earlier than Matlab 2018a.

% $Id: ReadAudioData.m,v 1.21 2020/11/18 22:44:31 pkabal Exp $

% Default buffer size
MAXBUFF = 50000;

% Tasks
ReadFromFile = 1;
DeleteBuffer = 2;
ReadFromBuffer = 3;
RefreshBuffer = 4;

% Decode the arguments
[File, AFPar, offs, Ns] = DecArg(varargin{:});

% Call PrintAudioPar if the first argument is a file name
if (~isempty(File))
  AFPar = PrintAudioPar(File);
  if (isfield(AFPar, 'CompressionMethod'))
    CheckCM(AFPar.CompressionMethod);
  end
end

Nframe = AFPar.TotalSamples;
Nchan = AFPar.NumChannels;
if (nargout < 3)
  NBuffMax = 0;
elseif (isfield(AFPar, 'NBuffMax'))
  NBuffMax = AFPar.NBuffMax;
else
  NBuffMax = MAXBUFF;
end

if (isempty(Ns))
  Ns = max(0, Nframe-offs);
end

% Find number of leading zeros and the number of samples taken from the file
[Nz, Nf, RFlim] = FindRFlim(Ns, offs, Nframe);

% See if a buffer has been allocated
if (isfield(AFPar, 'Buff'))
  NBuff = size(AFPar.Buff, 1);
  BFlim = [AFPar.Offs, AFPar.Offs+NBuff-1];   % Buffer offset limits
else
  BFlim = [0, -1];      % No buffer (0 length)
end

% Buffering is used if the AFPar output argument is present
% - The buffer is a field added to the AFPar struncture.
% - The buffer is deleted if the requested file data is larger than the maximum
%   buffer size.
% - The buffer is deleted if no samples from the file are present in the output.
% - The buffer is refreshed if not all samples can be obtained from the buffer.
% - There is an inherent inefficiency in copying the buffer into this routine
%   and then copying it back out.

% Allocate storage and zero the output array
x = zeros(Ns, Nchan);

if (Nf <= 0)
  % No file data needed
  Tasks = DeleteBuffer;
elseif (Nf > NBuffMax)
  % Read data directly into x
  Tasks = [DeleteBuffer, ReadFromFile];
elseif (RFlim(1) >= BFlim(1) && RFlim(2) <= BFlim(2))
  % All requested data available in the buffer
  Tasks = ReadFromBuffer;
else
  % Refresh buffer, move data from the buffer to the output
  Tasks = [RefreshBuffer, ReadFromBuffer];
end

% Carry out Tasks
for Task = Tasks
  switch Task
    case ReadFromFile
      x(Nz+1:Nz+Nf,:) = audioread(AFPar.Filename, RFlim+1);
    case DeleteBuffer
      if (BFlim(2) >= BFlim(1))
        AFPar = rmfield(AFPar, {'Buff', 'Offs'});
      end
    case ReadFromBuffer
      BBlim = RFlim - BFlim(1);
      x(Nz+1:Nz+Nf,:) = AFPar.Buff(BBlim(1)+1:BBlim(2)+1,:);
    case RefreshBuffer
      BFlim = UpdateBFlim(RFlim, BFlim, Nframe, NBuffMax);
      AFPar.Buff = audioread(AFPar.Filename, BFlim+1);
      AFPar.Offs = BFlim(1);
  end
end

% Number of samples, excluding trailing zero padding
Nv = Nz + Nf;

end

% ----- -----
function [Filename, AFPar, offs, Ns] = DecArg(varargin)

% Test for non-empty character vector
isCharVector = @(s) (ischar(s) && isrow(s));
% Test for a non-empty numeric scalar
isNumericScalar = @(x) (isnumeric(x) && isscalar(x));
% Test for a scalar structure
isStructScalar = @(x) (isstruct(x) && isscalar(x));

if (~isempty(varargin) >= 1 && isCharVector(varargin{1}))
  Filename = varargin{1};
  varargin(1) = [];
  AFPar = [];
elseif (~isempty(varargin) >= 1 && isStructScalar(varargin{1}))
  AFPar = varargin{1};
  varargin(1) = [];
  Filename = [];
else
  error('ReadAudioData: Invalid first argument');
end

offs = 0;
Ns = [];
if (~isempty(varargin) && isNumericScalar(varargin{1}))
  offs = varargin{1};
  varargin(1) = [];
end
if (~isempty(varargin) && isNumericScalar(varargin{1}))
  Ns = varargin{1};
  varargin(1) = [];
end

if (~isempty(varargin))
  error('ReadAudioPar: Too many input arguments');
end

end

% ----- -----
function [Nz, Nf, RFlim] = FindRFlim(Ns, offs, Nframe)

% Number of zeros before the start of file data
Nz = 0;
if (offs < 0)
  Nz = min(Ns, -offs);
end

% Initial RFlim
RFlim = [max(offs,0), min(Nframe-1,offs+Ns-1)];
Nf = RFlim(2) - RFlim(1) + 1;
if (Nf < 0)
  Nf = 0;
  RFlim = [];
end

end

% ----- -----
function CheckCM(CM)
% Check the compression method
% A-law and mu-law are broken in audioread before R2018b (9.5)

% Log-PCM in audioread was fixed in Matlab 2018a
if ((strcmp(CM, 'A-law') || strcmp(CM,'u-law')) && verLessThan('matlab', '9.5'))
  error('ReadAudioData: Incorrect %s data, fixed in Matlab 2018a');
end

end

% ----- ------
function BFlim = UpdateBFlim(RFlim, BFlim, Nframe, NBuffMax)

% File data will be read into the buffer
% If space permits, the buffer contains read-ahead or read-behind data. If the
% buffer has not been allocated yet, the bias will be to favour read-ahead. If
% the data offset is moving forward, we use read-ahead, and if the data offset
% is moving backward, we use read-behind.
if (BFlim(2) < BFlim(1))
  % No current buffer
  BFlim(2) = min(RFlim(1) + NBuffMax - 1, Nframe - 1);
  BFlim(1) = max(0, BFlim(2) - NBuffMax + 1);

elseif (RFlim(1) >= BFlim(1))
  % Moving forward
  NBuff = min(Nframe-1-RFlim(1)+1, NBuffMax);
  BFlim = [RFlim(1), RFlim(1)+NBuff-1];

else
  % Moving backward
  NBuff = min(RFlim(2)+1, NBuffMax);
  BFlim = [RFlim(2)-NBuff+1, RFlim(2)];

end

end
