function LSFC = LPxLSFQ(a, lsfQP, LSFpar)
% Convert LP parameters to LSFs, quantize the LSFs.

% $Id: LPxLSFQ.m,v 1.5 2020/11/19 02:21:07 pkabal Exp $

% Convert to LSFs
ap = a .* LSFpar.ECWin;
lsf = poly2lsf(ap);   % lsf's are in radians

% Quantize the LSFs
LSFC = QLSF(lsf, lsfQP, LSFpar);

end

%--------------------
function LSFC = QLSF(lsf, lsfQP, LSFpar)
% Quantize LSFs
% - Weighting inversely proportional to LSF spacings
% - Form the LSFs less the mean LSFs
% - Calculate the prediction error from the LSFs from the
%   previous frame

% Form the weight vector
% W(i) = 1 / min(lsf(i+1)-lsf(i), lsf(i)-lsf(i-1))
Slsf = diff(lsf);
W = 1 ./ min([Slsf(1); Slsf], [Slsf; Slsf(end)]);

% Calculate the prediction error
Pval = LSFpar.Pcof * (lsfQP - LSFpar.Mean);
Dlsf = (lsf - LSFpar.Mean) - Pval;

IQ = SVQLSF(Dlsf, W, LSFpar.VQ);
LSFC = IQ - 1;

end

%--------------------
function IQ = SVQLSF(Dlsf, W, YQ)
% Returns a vector of codebook indices

Nsplit = length(YQ);

i1 = 1;
IQ = NaN(1, Nsplit);
for k = 1:Nsplit
  i2 = i1 + size(YQ{k}, 1) - 1;
  IQ(k) = VQ(Dlsf(i1:i2), W(i1:i2), YQ{k});
  i1 = i2 + 1;
end

end

%--------------------
function Index = VQ(x, W, YQ)
% Returns the codebook index minimizing the weighted error

% We want to minimize (where W is a diagonal matrix)
%   (x-y)'W (x-y) = x'Wx - 2 y'W x + y'W y
%                 = x'Wx + y'W (y - 2 x)

Ny = size(YQ, 2);
ErrMin = inf;
for k = 1:Ny
  Err = (W .* YQ(:,k))' * (YQ(:,k) - 2 * x);
  if (Err < ErrMin)
    ErrMin = Err;
    Index = k;
  end
end

end
