function ProgressBar(Frac, Delta, Title)
% ProgressBar(Frac, Delta, Title)   % Initialization
% ProgressBar(Frac, Delta, Title)
% ProgressBar(Frac)                 % Update the progress bar
% Progress()                        % Close the progress bar
% Frac  -> Progress fraction (0 to 1). For initialization, this is the starting
%          progress (default 0);
% Delta -> Increment (0 to 1) between updates to the progress bar window. The
%          progress bar window is updated every time the progress crosses a
%          multiple of Delta. For instance if Delta is 0.05, the progress is
%          updated when the progress crosses 5%, 10%, ... . If Delta is small
%          enough, the progress bar window is updated for every update call. The
%          default value for Delta is 0.01, which implies at most 100 updates to
%          the visible progress bar (assuming the progress values are
%          monotonically increasing).
% Title -> Title of the progress bar window. The default is no title.
%
% The progress bar shows the progress of a computation with an annotation of
% progress in percent. This routine is a wrapper around the Matlab function
% waitbar. Alongside the bar, it prints the progress in percent. The granularity
% between updates to the bar can be set. For instance if the granularity is set
% to 5%, the bar will only be updated when the progress crosses a 5% boundary.
% This reduces the computational load for situations in which there are many
% small updates to the progress.
%
% Example:
%  ProgressBar(0, 0.02, 'Progress ...');
%  for i = 1:M
%    ProgressBar((i-1)/(M-1)); % Update bar if appropriate
%    ...
%  end
%  ProgressBar();        % Close the progress bar

% $Id: ProgressBar.m,v 1.3 2020/11/03 18:46:28 pkabal Exp $

persistent nsFrac h sDelta

if (nargin > 1)

  % Initialization - close waitbar if it is open
  if (~isempty(h) && isvalid(h))
    close(h);
  end

  % Resolve Delta / Title
  if (nargin == 2)
    if (ischar(Delta))
      Title = Delta;
      Delta = 0.01;       % Default Delta value
    else
      Title = '';         % Default Title
    end
  end
  sDelta = Delta;

  % Start up the progress bar
  Text = sprintf('%2.0f%% ...', 100 * Frac);
  h = waitbar(Frac, Text, 'Name', Title, 'Color', [1 1 .875]); % LightYellow

  % Set the initial nsFrac value
  nsFrac = floor(Frac / sDelta);

elseif (nargin == 1)
  if (isempty(h))
    error('ProgressBar: Progress bar has not been initialized');
  end

  % Update the progress
  nFrac = floor(Frac / sDelta);
  if (nFrac ~= nsFrac)
    Text = sprintf('%2.0f%% ...', 100 * Frac);
    waitbar(Frac, h, Text);
    nsFrac = nFrac;
  end

else

  % Close the waitbar
  if (~isempty(h))
    close(h);
  end
  h = gobjects(0);
end

end
