function RunRefCoder(FNameIn, FNameBit, Options)
% Run the reference ITU-G.723.1 coder
% FNamneIn -> Input audio file name (8 kHz sampling rate, single channel)
% FNameBit -> Output bit stream file name. This bit stream file can be decoded
%             by the G.723.1 decoder.
% Options  -> Operational mode: '-r53' (ACELP mode) or '-r63' (Multipulse mode)
%
% The data from the input file is written to a temporary headerless file with
% 16 bit integer data in little-endian order. This headerless file is input to
% the (floating point) reference coder. The reference coder has been compiled
% with Cygwin under Windows.

% $Id: RunRefCoder.m,v 1.2 2020/12/08 14:00:53 pkabal Exp $

LBCCODE = fullfile('bin', 'LBCCODE2.exe');
if (~ispc() || ~exist(LBCCODE, 'file'))
  error('RunRefCoder: Reference Codec not available');
end

% Throw an error if the file is not a recognized audio file
AFpar = PrintAudioPar(FNameIn);
if (AFpar.SampleRate ~= 8000 || AFpar.NumChannels ~= 1)
  error('RunRefCoder: Invalid file, expected single channel 8kHz sampling');
end

% Convert the double data to 16-bit resolution (with saturation)
x = int16(round(32768 * ReadAudioData(AFpar)));

% Temp file name
[~, Name] = fileparts(FNameIn);
FNameNH = [Name '.nh'];

% Write to the temporary file as native Windows (little-endian) data
FID = fopen(FNameNH, 'wb');
if (FID == -1)
  error('Cannot open file: %s', FullName(FNameNH));
end
fwrite(FID, x, 'integer*2');
fclose(FID);

if (nargin < 3)
  Options = '-r63';
end

% Flags
%  -r53 or -r63 or -rFilename: rate or rate file name (1 byte per frame)
%  -c, -d, -b: coder only, decoder only, or both
%  -v: use VAD/CNG
%  -n: Suppress printing
%  -Noh: No highpass filter
%  -Nop: No postfilter
%  -Rxx: Reinit size (reinitialize coder/decoder every xx frames)
%  -fFile: Frame erasure file (1 byte per frame)
LBCCODE = fullfile('bin', 'LBCCODE2.exe');
eval(['! ' LBCCODE ' -n -c ' Options ' ' FNameNH ' ' FNameBit]);

fprintf('G.723.1 Bitstream file: %s\n', FullName(FNameBit));

% Delete the temporary file
delete(FNameNH);

end
