function RunRefDecoder(FNameBit, FNameOut, FNameErr)
% Run the ITU-T G.723.1 reference decoder
% FNameBit -> Input bit stream file name (from a G.723.1 coder)
% FNameOut -> Output audio file name. This name should have the extension
%             '.wav'. The file will have 16 bit integer data at a 8 kHz sampling
%             rate.
% FNameErr -> Bitstream frame erasure file (1 byte per frame). Omit this
%             argument or set it to empty for no frame erasures.
%
% This function uses the reference (floating point) G.723.1 decoder compiled
% with Cygwin under Windows. The reference G.723,1 decoder processes the input
% bit file to produce a temporary headerless file containing 16 bit integer data
% in little-endian byte order. The data from the headerless file is read and
% rewritten to a WAV file.

% $Id: RunRefDecoder.m,v 1.4 2020/12/08 14:04:08 pkabal Exp $

LBCCODE = fullfile('bin', 'LBCCODE2.exe');
if (~ispc() || ~exist(LBCCODE, 'file'))
  error('RunRefDecoder: Reference Codec not available');
end

[~, Name] = fileparts(FNameBit);
FNameCnh = [Name '_Refdec.nh'];

% Flags
%  -r53 or -r63 or -rFilename: rate or rate file name (1 byte per frame)
%  -c, -d, -b: coder only, decoder only, or both
%  -v: use VAD/CNG
%  -n: Suppress printing
%  -Noh: No highpass filter
%  -Nop: No postfilter
%  -Rxx: Reinit size (reinitialize coder/decoder every xx frames)
%  -fFile: Frame erasure file (1 byte per frame)

if (nargin == 2 || isempty(FNameErr))
  FErr = [];
else
  FErr = ['-f' FNameErr ' '];
end
eval(['! ' LBCCODE ' -n -d ' FErr FNameBit ' ' FNameCnh]);

% Read the data from the headerless file
FID = fopen(FNameCnh);
if (FID == -1)
  error('Cannot open headerless file file: %s', FullName(FNameCnh));
end
x = fread(FID, 'int16=>double') / 32768;
fclose(FID);
delete(FNameCnh);

% Write the data to an audio file, add the date to the comment
DateStr = datestr(now, 31);
WriteAudioData(x, FNameOut, 8000, 16, 'Comment', ...
               'Floating-Point Reference Decoder');

end
