function status = CompAudioData(FNameA, FNameB, Delay)
% Print audio file statistics and compare samples in two audio files.
% status = CompAudioData(FNameA, FNameB, Delay)  % Delay samples in File B
%          CompAudioData(FNameA, FNameB)         % Delay = 0
%          CompAudioData(FNameA)                 % Statistics of File A
% status <- Status. The status value is set to 1 for a single file or when two
%           files return the same data. It is set to 0 if the data in two files
%           are different.
% FNameA -> Input file name, reference data for comparisons. If this is the only
%           argument, the statistics for the data in this file are printed.
% FNameB -> Input file name
% Delay  -> Vector of delay values. Each delay value representsa delay of data
%           in File B relative to the data in File A. A negative delay is a
%           time advance. If this argument is absent, the Delay is 0. Multiple
%           of delay are evaluated to find the delay which gives the largest
%           gain-optimized signal-to-noise ratio. Detailed file comparisons are
%           carried out for the best delay.
%
% For SNR comparisons between multi-channel audio files, the data are treated as
% a single channel. For stereo files, the SNR calculated is the same as if the
% two channels contain the real and imaginary components of complex samples.
%
% For each file, the following statistical quantities are calculated and
% printed.
%  Mean:
%       Xm = SUM x(i) / N
%  Standard deviation:
%       sd = sqrt[ (SUM x(i)^2 - Xm^2) / (N-1) ]
%  Max value:
%       Xmax = max(x(i))
% Min value:
%       Xmin = min(x(i))
% The above values are reported as a percent of full scale. For instance for a
% file with 16-bit integer data, full scale is 32768.
%
% An optional delay range can be specified when comparing files. The samples in
% file B are delayed or advanced relative to those in file A by each of the
% delay values in the delay range. For each delay, the SNR with optimized gain
% factor (see below) is calculated. For the delay corresponding to the largest
% SNR (with optimized gain), the full regalia of file comparison values is
% reported.
%
% In calculating the SNR for stereo files, the channels are considered to
% represent the real and imaginary parts of a complex signal. For multi-channel
% files, the samples in the files are considered to be a single channel.
%
% Conventional SNR:
%                 SUM |xa(i)|^2
%        SNR = --------------------- .
%              SUM |xa(i) - xb(i)|^2
%     The corresponding value in dB is printed.
%
% SNR with the optimized gain factor:
%       SNR = 1 / (1 - r^2) ,
%     where r is the (normalized) correlation coefficient,
%                     SUM xa(i)*xb'(i)
%       r = ------------------------------------- .
%           sqrt[ SUM |xa(i)|^2 * SUM |xb(i)|^2 ]
%     The SNR value in dB is printed. This SNR calculation corresponds to using
%     an optimized gain factor Sf for file B,
%            SUM xa(i)*xb'(i)
%       Sf = ---------------- .
%             SUM |xb(i)|^2

% $Id: CompAudioData.m,v 1.6 2020/12/08 03:47:31 pkabal Exp $

dBP = @(P) (10*log10(P));
SAME = 1;
DIFFERENT = 0;

if (nargin == 1)
  NFiles = 1;
else
  NFiles = 2;
end
if (nargin <= 2)
  Delay = 0;
end

AFparA = PrintAudioPar(FNameA);
if (NFiles == 2)
  AFparB = PrintAudioPar(FNameB);
  if (AFparA.NumChannels ~= AFparB.NumChannels)
    error('CompAudioData: Files do not have the same number of channels');
  end
  if (AFparA.TotalSamples ~= AFparB.TotalSamples)
    fprintf('CompAudioData: No. samples differ, using maximum\n');
  end
end
if (AFparA.NumChannels > 1)
  fprintf('CompAudioData: Multichannel data, treating as 1 channel\n');
end

% Individual file statistics and cross-file statistics
FA = FileStats(AFparA);
if (NFiles == 2)
  FB = FileStats(AFparB);
  FAB.SNRG = -Inf;
  DelayOpt = Delay(1);
  for delay = Delay
    FABt = CrossStats(AFparA, AFparB, delay);
    if (FABt.SNRG > FAB.SNRG)
      DelayOpt = delay;
      FAB = FABt;
    end
    if (length(Delay) > 1)
      fprintf(' Delay = %d, SNR = %.3g dB (File B gain = %.2f)\n', ...
              delay, dBP(FABt.SNRG), FABt.GainB); 
    end
  end
end

% Print individual file statistics
if (NFiles == 2  && FAB.NDiff > 0)
  fprintf(' File A:\n');
  fprintf('   Std Dev: %.3g%%, Mean: %.3g%%\n', 100*FA.StD, 100*FA.Mean);
  fprintf('   Max: %.3g%%, Min: %.3g%%\n', 100*FA.Max, 100*FA.Min);
  fprintf(' File B:\n');
  fprintf('   Std Dev: %.3g%%, Mean: %.3g%%\n', 100*FB.StD, 100*FB.Mean)
  fprintf('   Max: %.3g%%, Min: %.3g%%\n', 100*FB.Max, 100*FB.Min);
else
  fprintf('   Std Dev: %.3g%%, Mean: %.3g%%\n', 100*FA.StD, 100*FA.Mean);
  fprintf('   Max: %.3g%%, Min: %.3g%%\n', 100*FA.Max, 100*FA.Min);
end

if (NFiles == 2)

  if (length(Delay) > 1)
    fprintf(' Best Match at delay = %d\n', DelayOpt);
  elseif (DelayOpt ~= 0)
    fprintf(' Delay = %d\n', DelayOpt);
  end
  
  if (FAB.NDiff == 0)
    fprintf(' File A = File B\n');
  elseif (isinf(FAB.SNRG))
    fprintf(' File A = %.2f * File B\n', FAB.GainB);
  else
    if (FAB.Nsamp ~= FA.Nsamp || FAB.Nsamp ~= FB.Nsamp)
      fprintf(' Number of samples: %d\n', FAB.Nsamp);
    end
    fprintf(' SNR = %.2f dB\n', dBP(FAB.SNR));
    fprintf(' SNR = %.2f dB (File B Gain = %.3f)\n', dBP(FAB.SNRG), FAB.GainB);
    if (AFparA.NumChannels == 1)
      fprintf(' Max Diff: %.3g%% full scale, No. Diff = %d (%d runs)\n', ...
              100*FAB.MaxADiff, FAB.NDiff, FAB.Nrun);
    else
      fprintf(' Max Diff: %.3g%% full scale, No. Diff = %d/%d\n', ...
              100*FAB.MaxADiff, FAB.NDiff, FAB.Nsamp);
    end

  end
end

status = SAME;
if (NFiles == 2 && FAB.NDiff > 0)
  status = DIFFERENT;
end

end

% ----- ------
function F = FileStats(AFpar)

x = ReadAudioData(AFpar, 0, AFpar.TotalSamples);
F.Mean = mean(x(:));
F.StD = std(x(:), 1);
F.Max = max(x(:));
F.Min = min(x(:));
F.Nsamp = length(x(:));

end

% ----- -----
function FAB = CrossStats(AFparA, AFparB, Delay)

Nsamp = AFparA.NumChannels * max(AFparA.TotalSamples, AFparB.TotalSamples);

% A delay of D samples means that an event that occurs at time n in, now occurs
% at time n+D.
% Alignment               pos D             neg D
%        0      N-1     0      N-1         D   0      N-1
%   Xa = xxxxxxxx    => xxxxxxxx....    => ....xxxxxxxx
%   Xb = yyyyyyyy    => ....yyyyyyyy    => yyyyyyyy....
%        0      N-1    -D   0      N-1     0      N-1
%
% The file data is conceptually extended with zeros at either end. The length of
% vectors after shifting is N+|D|. If |D| is greater than Nsamp, the effective
% delay can be set to +/-Nsamp - there will be no overlap of Xa and Xb.

Edelay = max(-Nsamp, min(Nsamp, Delay));     % -Nsamp <= delay <= Nsamp
FAB.Nsamp = Nsamp + abs(Edelay);
offs = min(0, Edelay);
Xa = ReadAudioData(AFparA, offs, FAB.Nsamp);
Xb = ReadAudioData(AFparB, offs - Edelay, FAB.Nsamp);

% Maximum absolute difference
e = Xa - Xb;
FAB.MaxADiff = max(abs(e(:)));
FAB.NDiff = sum(e(:) ~= 0);

% Calculate the number of error runs
Nrun = 0;
if (AFparA.NumChannels == 1)
 Nrun = sum(diff(e(:) ~= 0) == 1);  % Start of run
  if (e(1) ~= 0)
    Nrun = Nrun + 1;
  end
end
FAB.Nrun = Nrun;

Sxx = Xa(:)' * Xa(:);
See = e(:)' * e(:);
Sxy = Xa(:)' * Xb(:);
Syy = Xb(:)' * Xb(:);

% Calculate the SNR's
FAB.SNR = Sxx / See;
FAB.GainB = Sxy / Syy;
denom = Sxx - Sxy * (Sxy / Syy);
FAB.SNRG = Sxx / denom;

end
