function FileName = FullName(FName)
% FileName = FullName(Fname)
%
% FileName <-  Full file name
% Fname     -> Partial file name for an existing file specified with a relative
%              or absolute path. It will also open files that are on the Matlab
%              path.
%
% Find the full file name for an existing file. This routine opens the file,
% extracts the full name, and closes the file. An error is signalled if the file
% does not exist. For Windows, there is a second step to clean up mismatched
% case in the path components and the file name and extension.

% $Id: FullName.m,v 1.8 2020/11/19 02:23:44 pkabal Exp $

% Note
% - On Windows, in general the path components, file name and extension are not
%   case sensitive matches, but ...
%   - The special folder name 'private' is case sensitive
%   - Consider the file specification 'abc/def.ghi'.
%      'Abc/Def.Ghi' is returned as '-path-\Abc\Def.Ghi'
%      'abc/Def.Ghi' is returned as '-path-\abc\Def.Ghi'
%   - For a file 'def.ghi' in the current folder, 'Def.Ghi' is returned as
%     '-path-\def.ghi' (the correct case)
% - On Windows, the path separator can be '/' or '\'. The returned full name
%   uses the '\' separator.

% Get the file name using fopen
FID = fopen(FName);
if (FID < 0)
  error('FullName: file not found');
end
FileName = fopen(FID);
fclose(FID);

% For Windows, clean up the mixed case match
if (ispc())
  FileName = FixFileName(FileName);
end

end

% ----- -----
function FileName = FixFileName(FileName)

% Go to the file folder to get a path with the correct case
[Path, Name, Ext] = fileparts(FileName);
CWD = cd(Path);
FileName = which([Name, Ext]);    % Clean up the case of the Name/Ext
cd(CWD);

end
