function AFPar = PrintAudioPar(Fname)
% Print information about an existing audio file.
% AFPar = PrintAudioPar(Fname)
%         PrintAudioPar(Fname)
% AFPar <-  Output structure containing information about the file
% Fname  -> File name (on the Matlab path, or with an explicit path)
%
% The output structure AFPar returns the information from audioinfo. An example
% of the format of the printout:
%
% Audio file: D:\MMSP\Matlab\Audio\test\test16.flac
%   Samples / channel : 93 (0.0093 s)
%   Sampling frequency: 10000 Hz
%   Number of channels: 2 (16-bit FLAC)
%
% The fields in the output structure AFPar are:
%   AFPar.Filename: File name, including the full path to the file. On Windows,
%      this routine modifies the FileName to return the file name in the correct
%      case.
%   AFPar.SampleRate: Sampling frequency for each channel (samples/sec)
%   AFPar.TotalSamples: Number of samples per channel
%   AFPar.NumChannels: Number of channels
%   AFPar.BitsPerSample: Number of bits per sample in the audio file
%   AFPar.BitRate: Bitrate in kb/s (for some compressed formats)
%   AFPar.Artist: Artist, empty if missing
%   AFPar.Title: Title, empty if missing
%   AFPar.Comment: Comment, empty if missing
%   AFPar.DataFormat: Data format. This field is added by this routine to the
%      structure returned by audioinfo. This is a printable character string
%      describing the data format. Examples: '8-bit mu-law', '32-bit float',
%      '32-bit integer'.

% $Id: PrintAudioPar.m,v 1.14 2020/12/04 17:13:02 pkabal Exp $

% Get information from the audio file
AFPar = audioinfo(Fname);

% Notes on the information structure returned by audioinfo:
% - The TotalSamples field is inappropriately named - it is the number of
%   samples per channel.
% - It is not possible, based on the information returned by audioinfo, to
%   distinguish 32-bit integer from 32-bit float data. This routine peeks inside
%   the file to determine the data type.
% - If there are multiple information records of a given type (comments for
%   instance) only the first one found is returned by audioinfo.
% - On Windows, the file name may be a case-insensitive match to the actual file
%   name.

% Fix up the case of Filename
if (ispc())
  AFPar.Filename = FixFileName(AFPar.Filename);
end

% Generate a printable data format string
AFPar.DataFormat = GenDFormat(AFPar);

% Print the file information
Nchan = AFPar.NumChannels;
Nframe = AFPar.TotalSamples;
Sfreq = AFPar.SampleRate;
fprintf(' Audio file: %s\n', AFPar.Filename);
fprintf('   Samples / channel : %d (%.4g s)\n', Nframe, Nframe/Sfreq);
fprintf('   Sampling frequency: %g Hz\n', Sfreq);
fprintf('   Number of channels: %d (%s)\n', Nchan, AFPar.DataFormat);

% Print information records
if (isfield(AFPar, 'Title') && ~isempty(AFPar.Title))
  fprintf('   Title: %s\n', AFPar.Title);
end
if (isfield(AFPar, 'Artist') && ~isempty(AFPar.Artist))
  fprintf('   Artist: %s\n', AFPar.Artist);
end
if (isfield(AFPar, 'Comment') && ~isempty(AFPar.Comment))
  fprintf('   Comment: %s\n', AFPar.Comment);
end


end

% ----- -----
function FileName = FixFileName(FileName)

% Go to the file folder to get a path with the correct case
[Path, Name, Ext] = fileparts(FileName);
CWD = cd(Path);
FileName = which([Name, Ext]);    % Clean up the case of the Name/Ext
cd(CWD);

end

% ----- -----
function DataFormat = GenDFormat(AFPar)

% Generate a name for the file data format

if (isfield(AFPar, 'BitsPerSample'))
  Nbit = AFPar.BitsPerSample;
else
  Nbit = NaN;
end
if (isfield(AFPar, 'BitRate'))
  BitRate = AFPar.BitRate;
else
  BitRate = NaN;
end

CM = AFPar.CompressionMethod;
switch CM
  case 'u-law'
    DataFormat = '8-bit mu-law';
  case 'A-law'
    DataFormat = '8-bit A-law';
  case 'Uncompressed'     % Integer (8,16,24,32 bits) or float (32,64 bits)
    if (Nbit < 32)
      DataFormat = sprintf('%d-bit integer', Nbit);
    elseif (Nbit == 32)
      y = audioread(AFPar.Filename, [1, 1], 'native');  % Peek inside the file
      switch class(y)
        case 'single'
          DataFormat = sprintf('%d-bit float', Nbit);
        case 'int32'
          DataFormat = sprintf('%d-bit integer', Nbit);
        otherwise
          DataFormat = sprintf('%d-bit uncompressed', Nbit);
      end
    elseif (Nbit == 64)
      DataFormat = sprintf('%d-bit float', Nbit);
    else
      DataFormat = sprintf('%d-bit uncompressed', Nbit); % integer or float
    end
  otherwise
    if (~isnan(Nbit))
      DataFormat = sprintf('%d-bit %s', Nbit, CM);
    elseif (~isnan(BitRate))
      DataFormat = sprintf('%d kb/s %s', BitRate, CM);
    else
      DataFormat = CM;
    end
end

end
