function CheckDiaryFile = StartDiaryFile(LogFile, RefFile, SysInfo)
% CheckDiaryFile = StartDiaryFile(LogFile, RefFile, SysInfo)
% CheckDiaryFile = StartDiaryFile(SysInfo)
% CheckDiaryFile = StartDiaryFile()
%
% CheckDiaryFile <-  Pointer to function to compare the log file to the
%                    reference file
% LogFile         -> Diary file containing the output text (default name
%                    determined from the module name)
% RefFile         -> Reference file containing past diary file output, default
%                    name determined from the module name
% SysInfo         -> Logical flag, if set, the diary file will have a header
%                    containing the Matlab version and the hostname of the
%                    computer. The default is off. This argument can be the last
%                    argument if either or both of the LogFile and RefFile
%                    arguments are missing.
%
% To check the correct functioning of a Matlab function, a test routine can be
% written to exercise the Matlab function with different test cases. The output
% of the test cases can be written to the screen and simultaneously to a diary
% file. StartDiarFile/CheckDiaryFile will determine whether changes to the
% Matlab function result in changes to the output of the test routine as stored
% in the diary file.
%
% Using StartDiaryFile, output from the test function is routed to the diary
% file (here called the log file). Using the companion routine CheckDiaryFile,
% the log file is compared with a reference log file. The reference file
% contains the expected output of the test script. StartDiaryFile supplies
% defaults for LogFile and RefFile based on the name of the routine calling this
% routine. If the calling function is testfn.m, the defaults are 'testfn.log'
% and 'testfn.ref'.
%
% StartDiaryFile:
% 1. Close any open diary files.
% 2. Start the diary file. The Matlab diary is sent to the log file (if the log
%    file name is not empty).
% 3. Save the current output format parameters. Set the output format to 'short'
%    and 'compact'.
% 4. Return a pointer to the routine CheckDiaryFile. The saved format parameters
%    and the file names for the log and reference files are available to
%    CheckDiaryFile.
%
% CheckDiaryFile:
% 1. Close the diary file.
% 2. Reset the output formatting to the settings prior to calling SetDiaryFile.
% 3(a) If the log file name is empty, no further action is taken.
%  (b) If the log file name differs from the reference file name, the contents
%      of these files are compared using the system file comparison tool.
%  (c) If the file contents are identical, the log file is deleted.
%
% Notes:
% - For a first run with no reference file, the log file will be created. That
%   log file can be renamed to become the reference file for subsequent runs.
% - When Matlab displays the contents of a matrix (disp command for instance),
%   the number of elements displayed per line depends on the width of the Matlab
%   command window. This means that a comparison of the log and reference files
%   may show differences when the numerical values are unchanged. The workaround
%   is to print matrices using formatted I/O.
%
% StartDiaryFile should be called at the beginning of the test script and
% CheckDiaryFile should be called at the end of the test script.
%   CheckDiaryFile = StartDiaryFile(LogFile, RefFile)
%   ...
%   CheckDiaryFile();
%
% Note that if the script fails before CheckDiaryFile turns diary off, the log
% file can be locked, which may require restarting Matlab to remove the lock.

% $Id: StartDiaryFile.m,v 1.16 2020/11/18 21:18:16 pkabal Exp $

% Process the arguments
switch nargin
  case 0
    LogFile = '';
    RefFile = '';
    SysInfo = false;
  case 1
    if (islogical(LogFile))
      SysInfo = LogFile;
      LogFile = '';
      RefFile = '';
    else
      RefFile = '';
      SysInfo = false;
    end
  case 2
    if (islogical(RefFile))
      SysInfo = RefFile;
      RefFile = '';
    else
      SysInfo = false;
    end
end

% Get calling module name
if ((isempty(LogFile) || isempty(RefFile)) && size(dbstack(), 1) > 1)
  ST = dbstack(1);            % Skip the first entry (this file)
  Cmodule = ST(1).name;
else
  Cmodule = 'CommandWindow';  % Called from the Command window
end

if (isempty(LogFile))
  LogFile = [Cmodule, '.log'];
end
if (isempty(RefFile))
  RefFile = [Cmodule, '.ref'];
end

diary off;
if (~isempty(LogFile))
  if (~isempty(ls(LogFile)))
    % May need to pause here to let "diary off" take effect
    delete(LogFile);  % If it exists, we don't want to append to it
  end
  diary(LogFile);
end

DFormat = get(0, 'Format');
DFormatSpacing = get(0, 'FormatSpacing');
format;         % Set to default format (short)
format compact  % Change format spacing to compact

% Header
if (SysInfo)
  [~, hostname] = system('hostname');   % May have trailing CR-LF
  fprintf('--- Matlab %s on %s ---\n', ...
          version('-release'), deblank(hostname));
end

CheckDiaryFile = @Check_DiaryFile;

% ----- begin nested function
function Check_DiaryFile()
% This function compares a log file with a reference file. It is the companion
% function to SetDiaryFile. The parameters are inherited from StartDiaryFile.
%
% 1. Close the diary.
% 2. Reset the output formatting to the settings prior to calling SetDiaryFile.
% 3(a) If the log file name is empty, no further action is taken.
%  (b) If the log file name differs from the reference file name, the contents
%      of these files are compared using the system file comparison tool.
%  (c) If the files have identical content, the log file is deleted.

diary off

% Restore formating
set(0, 'Format', DFormat);
set(0, 'FormatSpacing', DFormatSpacing);

if (~isempty(ls(RefFile)) && ~strcmp(LogFile, RefFile))
  fprintf('\n***** ***** ***** *****\n');

  % System file compare
  if (ispc())
    Status = system(['FC ', LogFile, ' ', RefFile]);
  else
    Status = system(['diff ', LogFile, ' ', RefFile]);
  end

  % Delete the log file if the files are identical
  if (Status == 0)
    delete(LogFile);
  end
end

end
% ----- end nested function

end
