function WriteAudioData(x, Fname, Fs, Nbit, varargin)
% Write values to a WAV or FLAC audio file (integer file data)
% WriteAudioData(x, Fname, Fs, Nbit, Options)
% WriteAudioData(x, Fname, Fs, Nbit)
%
% x     -> Double-precision or single-precision real array dimensioned Ns x M,
%          where M is the number of channels and Ns is the number of samples per
%          channel. For single channel data, the data can be stored as either a
%          column or a row. The input is assumed to be normalized so that the
%          values fall in the nominal range -1 to +1.
% Fname -> File name. The file extension must be '.wav' or '.flac'.
% Fs    -> Sampling frequency for each channel (samples/second)
% Nbit  -> Number of bits per sample (8/16/24) in the WAVE file.
% Options -> Name/Value pairs for additional options to audiowrite. This can be,
%          for intance Title, Artist, or  Comment options.
%
% After writing the data, the parameters of the output file are printed out.
%
% WAV files
% - For float input data and WAV files with integer data, audiowrite prior to
%   Matlab 2015a scaled the input data incorrectly.
% - For float input data and WAV files with an odd number of total data bytes,
%   audiowrite prior to Matlab 2017b does not properly pad out the data
%   structure in the file.
% - There is a continuing problem with audiowrite - it scales the float data and
%   then converts the values to integers using the floor function. This routine
%   changes the behaviour by quantizing the scaled input float data using a
%   round function so that the resuting integer values that appear in the WAV
%   file are those that would occur if the round function were substituted for
%   the floor function. Use of rounding reduces the maximum magnitude of the
%   quantization error by a factor of two compared to use of the floor function.
%
% FLAC files
%   For float input data and FLAC files, audiowrite implementation implements a
%   variant of rounding of the scaled input data. Conventional rounding finds
%   the nearest integer value, with mid-points between integer values being set
%   to the largest magnitude neighbour. The rounding used in FLAC is banker's
%   rounding (also known as round half to even), with mid-points between integer
%   values being set the even neighbour. This routine does not change this
%   behaviour.
%
% Clipping Messages
%   Clipping occurs when the input data is outside the range that can be
%   represented by the integer data in the file. The clipping message generated
%   by audiowrite is based on the input float data being outside the range
%   -1 <= x <= +1. A value of +1 cannot be represented by the integer data in
%   the file, but an input of this value is clipped with no warning message.
%
%   For the default behaviour of audiowrite (using a floor function), the
%   correct check for clipping is
%     -1 <= floor(x) < +1.
%   For the round function used in this routine for WAV files, the correct check
%   for clipping is
%     -1-delta/2 < x < +1-delta/2,
%   where delta is the step size 1/2^(Nbit-1). For the banker's round function
%   used for FLAC files, the correct check for clipping is
%     -1-delta/2 <= x < +1-delta/2.
%   This routine termporarily turns off Matlab clipping warnings and prints its
%   own clipping message based on the comparisons noted above.

% $Id: WriteAudioData.m,v 1.13 2020/11/18 22:44:48 pkabal Exp $

FWAV = 1;
FFLAC = 2;

% x must be real and float (single or double)
if (~(isfloat(x) && isreal(x) && ismatrix(x)))
  error('WriteAudioData: Input data must be real double or real single matrix');
end

% Require wave/flac extension
[~, ~, ext] = fileparts(Fname);
if (strcmpi(ext, '.wav'))
  FileType = FWAV;
elseif (strcmpi(ext, '.flac'))
  FileType = FFLAC;
else
  error('WriteAudioData: Unsupported file type');
end

% For WAV files in audiowrite, if Nbit=32, the data is written as 32-bit floats
% and if Nbit==64, the data is written as 64-bit floats.
if (~(Nbit == 8 || Nbit == 16 || Nbit == 24))
  error('WriteAudioData: Only 8/16/24 bit integer file data is supported');
end

% Write the audio data
% - Prior to Matlab 2015a, audiowrite does not work properly for writing integer
%   data to WAV files). It uses an incorrect scaling: for instance for 16-bit
%   files and double x, x is scaled by 32767 instead of 32768).
% - Prior to Matlab 2017b, for odd length data
SV = 2^(Nbit-1);
if (FileType == FWAV)
  if (verLessThan('matlab', '8.5'))   % < Matlab 2015a
    error('WriteAudioData: Invalid WAV file scaling, fixed in Matlab 2015a');
  end
  DataLen = Nbit/8 * numel(x);
  if (mod(DataLen, 2) == 1 && verLessThan('matlab', '9.3')) % < Matlab 2017b
    warning('WriteAudioData - Malformed WAV file, fixed in Matlab 2017b');
  end
  xs = round(SV*x);
  x = xs / SV;  % Rounding
else
  xs = bankers(SV*x);
end
Nclip = sum(xs(:) < -SV | xs(:) >= SV);

% Turn off data clipped messages
SAWc = warning('off', 'MATLAB:audiovideo:audiowrite:dataClipped');

% Write the data to the file
audiowrite(Fname, x, Fs, 'BitsPerSample', Nbit, varargin{:});

warning(SAWc);      % Restore the warning

% Print the audio file parameters
PrintAudioPar(Fname);
if (Nclip > 0)
  fprintf('WriteAudioData - %d data values clipped\n', Nclip);
end

end

% ----- -----
function y = bankers(x)
% Rounding to nearest integer with half values to nearest even integer

y = round(x);
I = (abs(y-x) == 0.5);
y(I) = 2 * round(x(I)/2);

end
