function tG7231Decoder(FNameData, FNameErr, FNameRefWav)
% Compare the output of the Matlab decoder with a reference output file. If the
% reference file name is set to '*', the reference output is generated by the
% reference floating-point C-code decoder.
% 
% Default mode:
%   tG7231Decoder()                   % Runs through 3 test cases
%
% Single test case:
%   tG7231Decoder(FNameData, FnameErr, FnameRefWav)
%     FNameData:   Bit/Data file, input to G.723.1 decoder
%     FNameErr:    BitError file, input to G.723.1 decoder, can be empty
%     FNameRefWav: Reference WAVE file. If this is '*', a temporary WAVE file is
%                  created using the reference C decoder.%
% Multiple test mode:
%   tG7231Decoder(FNames)   % Cell array of file names, N x 3

% $Id: tG7231Decoder.m,v 1.13 2020/12/08 12:46:29 pkabal Exp $

SavePath = addpath('..');

if (nargin < 3)
  FNameRefWav = '';
end
if (nargin < 2)
  FNameErr = '';
end
if (nargin == 0)
  TDir = 'TestFiles';
  FNameData = ...
    {fullfile(TDir, 'FE27_04_Mref63.dat'), '', ...
                            fullfile(TDir,'FE27_04_Mref63.wav');
     fullfile(TDir,'FE27_04_Mref63.bit'), '', ...
                            fullfile(TDir,'FE27_04_Mref63.wav');
     fullfile(TDir, 'MA01_02_Cref53.bit'), '', ...
                            fullfile(TDir,'MA01_02_Cref53.wav')};
  SysOS = true;
  CheckDiaryFile = StartDiaryFile(SysOS);
end

if (nargin <= 1 && iscell(FNameData))
  NTest = size(FNameData, 1);
  Nequal = 0;
  for i = 1:NTest
    Nequal = Nequal + tDecoder(FNameData{i,1}, FNameData{i,2}, FNameData{i,3});
  end
else
  tDecoder(FNameData, FNameErr, FNameRefWav);
end

if (NTest > 0)
  fprintf('\nNumber of matched files: %d / %d\n', Nequal, NTest);
end

if (nargin == 0)
  CheckDiaryFile();
end
path(SavePath);

end

% ----------
function Same = tDecoder(FNameData, FNameErr, FNameRefWav)

[~, Name] = fileparts(FNameData);
FNameMwav = [Name '_Mdec.wav'];       % Temporary file

% Open the file for a peek at the first 6 characters
fid = fopen(FNameData);
if (fid == -1)
  error('G7231Decoder: Cannot open input file');
end
id = char(fread(fid, 6, 'uint8')');
fclose(fid);

isBitStream = (~strcmp(id, 'MATLAB'));

fprintf('----- ----- ----- ----- ----- -----\n');
% Run the Matlab decoder
fprintf('====== Running the Matlab decoder\n');
G7231Decoder(FNameData, FNameMwav, FNameErr);

% Run the reference C decoder
TempRefWav = (isBitStream && strcmp(FNameRefWav, '*'));
if (TempRefWav)
  [~, Name] = fileparts(FNameData);
  FNameRefWav = [Name '_Cdec.wav'];
  fprintf('\n====== Running the reference C decoder\n');
  RunRefDecoder(FNameData, FNameRefWav, FNameErr);
end

% Compare the output
if (~isempty(FNameRefWav))
  fprintf('\n====== Comparing WAVE files\n');
  Same = CompAudioData(FNameRefWav, FNameMwav);
end

% Clean up
if (TempRefWav)
  delete(FNameRefWav);
end
delete(FNameMwav);

end
